
/**
	Changelog:

	* Version 0.6 (BDF6)
		- Max tracker cap

	* Version 0.5.1 
		- Better handling of few inliers scenario
		- Small refactoring

	* Version 0.5
		- Support to initialization with a polygon (for a rotated rect)

	* Version 0.4:
		- Context-aware trackers
		- context trackers rember if they were transitioned to inliers. If they were not, they will never be reinitialized

	* Version 0.3:
		- textureless patch filtering

	* Version 0.2:
		- adaptive/multisize tracker patch size
		- new state: transition (when many trackers are disabled, reinitialize them with a transition state)
		- give weights to trackers with past good performance

	* Version 0.1:
		- base system design

	Todo:
		- ratio between inlier/outlier should not count context trackers

*/



using namespace std;


class LocalTracker {
	public:
		
		Point2i tl;				// Top-Left corner
		Point2i prevTL;			// previous position point
		

		int state;				// 1 => inlier; 2 => outlier; 3 => transition 

		Point2i displacement;	// Difference with the previous position

		int goodEstimate;		// Number of correctly tracked frames (used for weighting)

		// Initial position
		Point2i relativeTL;		// Relative position inside the bounding box
		int patchSize;			// Size of the patch

		bool context;			// Context support tracker or object

		int outlierAt = -10;			// Frame where the tracker was switched to outlier
		int inlierCount;		// Number of times this tracker was transitioned to inlier state (this is a good indicator for good context location)

		int transitionThreshold = -1;

		LocalTracker(Point2i relativeTL, int size, bool context = false) {

			this->relativeTL = relativeTL;
			this->patchSize = size;
			this->context = context;

			this->inlierCount = 0;
		}


		void setStateInlier() {
			state = 1;

			this->inlierCount++;
		}

		void setStateOutlier(int frame = 0) {
			//if (state == 1)
				outlierAt = frame;

			state = 2;
			
			this->goodEstimate = 0;
		}

		void setStateTransition() {
			state = 3;

			this->goodEstimate = 0;
		}

		// Set the old position to current position, and update the position
		void updateTL(Point2i tl) {
			
			this->prevTL = this->tl;
			this->tl = tl;

			this->displacement = this->tl - this->prevTL;

		}

		// Initialize the tracker in its default position and set it's state
		void reinit(const Mat& currentFrame, const Rect& objectROI, int setState = 1, int currentTransitionThreshold = -1) {

			this->goodEstimate=0;
			this->transitionThreshold = currentTransitionThreshold;
			
			Point2i tl = objectROI.tl() + this->relativeTL;
			updateTL(tl);
			
			Rect patchROI( tl.x, tl.y, this->patchSize, this->patchSize );

			
			// Disable current patch if: (i) it's out of the image or (ii) it's out of the bounding box
			if (!rectInsideImageCheck(currentFrame, patchROI) /*|| !objectROI.contains(this->tl)*/) { 
				
				setStateOutlier();
				return;	

			} 
			
			// If this is a context tracker, it should be initialized to transition state
			if (this->context==true)
				setStateTransition();
			else
				this->state = setState;
			

			

			// Check if the current patch has some edges
			/*
			Mat currentPatch = currentFrame(patchROI);
			uchar previous = currentPatch.at<uchar>(0,0);
			int diff = 0;
			for(int i=0;i<currentPatch.rows;i++)
				for(int j=0;j<currentPatch.cols;j++) {
					int current = currentPatch.at<uchar>(i,j);
					diff += abs(previous - current);
					previous = current;
				}
			int meanPixelDifference = 1;
			if ( diff < patchSize*patchSize*meanPixelDifference ) {
				setStateOutlier();
				return;
			} 
			//cout << "Tracker intensity difference: "<<diff<<", threshold: "<<patchSize*patchSize*meanPixelDifference<<endl;
			//showImage(frame);
			//waitKey(0);
			*/
	
		}

		void draw(Mat& frame) {
			// Default inlier green
			Scalar color(0,255,00);

			// Transition yellow
			if (state==3)
				color = Scalar(0,255,255);

			// Outlier red
			if (state == 2)
				color = Scalar(0,0,255);

			// Context blu
			if (this->context==true)
				color = Scalar(255,0,0);

			Rect patchROI( tl.x, tl.y, this->patchSize, this->patchSize );

			if (state != 2) {
				drawRect(frame, patchROI, color, 1);
				//drawRectCenter(frame,patchROI,color);
			}
		}
};

// LocalTracker to String
ostream& operator<<(ostream& strm, const LocalTracker& a) {
	return strm << " [Tracker] tl: " << a.tl <<  
		", prevTL: "<<a.prevTL<<
		", displacement: "<<a.displacement<<
		", size: "<<a.patchSize<<
		", state: "<<a.state<<
		", context: "<<(int)a.context<<
		", weight: "<<a.goodEstimate;
}

/*
// Reinit all trackers with state outlier ("outlier"->"transition")
void disabledTrackersToTransition(const Mat& currentFrame, const Rect& objectROI) {

	for(auto i=0;i<trackers.size();++i) {
		if (trackers[i].state == 2) {

			// Do not transition if a tracker is context and was never good
			if (trackers[i].context == true && trackers[i].inlierCount == 0)
				continue;

			int state = 3;
			// If we disabled transitionThreshold put it directly as inlier
			if (transitionThreshold==0)
				state = 1;

			trackers[i].reinit(currentFrame,objectROI,state);			// Set transition mode

			
		}
	}

}
*/

class BDF : public Tracker {

public:
	void track() {

		// Prepare for tracking (color conversions)
		this->prepare();

		//cv::Ptr<cv::CLAHE> clahe = cv::createCLAHE();
		//clahe->apply(currentFrame, currentFrame);


		// Init
		if (frameCount == 1) {

			objectROI = firstBB;
			if (corners.empty())
				corners = rectToCorners(firstBB);

			//initTrackers();
			initMultiSizeTrackers();

			Log(0) << "Total final trackers: " << trackers.size() << endl << endl;

			if (this->debug)
				showImage(frame, "Tracker");

			//waitKey(0);

			currentFrame.copyTo(prevFrame);
			return;
		}


		//   X		  Y  => Bucket
		map<int, map<int, int> > hs;
		map<string, size_t> trackerIndex;


		int maxVote = 0;
		Point2i bestDisplacement;

		vector<int> displacementX;
		vector<int> displacementY;

		// Check if a tracker is inside this area
		Rect objectROIEnlarged = rectScale(objectROI, 1.5f);

		//if (this->debug)
			//drawRect(frame, objectROIEnlarged, Scalar(255, 0, 0));

		// Displacements estimation
		for (auto i = 0; i<trackers.size(); i++) {

			auto& t = trackers[i];	// Current Tracker


			// If the current state is oulier
			if (t.state == 2) {		
				continue;			// Skip local motion estimation
			}


			// The tracker is outside the object
			if (!objectROIEnlarged.contains(t.tl)) {
				t.setStateOutlier();
				continue;
			}


			// Current tracker search window
			Rect searchWindow(
				t.tl.x - patchEnlargeWindow,
				t.tl.y - patchEnlargeWindow,
				t.patchSize + (2 * patchEnlargeWindow),
				t.patchSize + (2 * patchEnlargeWindow)
			);

			// Check if the search window is inside the image
			if (!rectInsideImageCheck(currentFrame, searchWindow))
				continue;


			// Search Window
			//drawRect(frame,searchWindow,Scalar(255,0,0));
			//showImage(frame);
			//waitKey(0);

			Rect currentPatchROI = Rect(t.tl.x, t.tl.y, t.patchSize, t.patchSize);
			Mat currentPatch = prevFrame(currentPatchROI);

			Mat result;
			matchTemplate(currentFrame(searchWindow), currentPatch, result, CV_TM_CCORR_NORMED);

			double minVal, maxVal; 	Point minLoc, maxLoc, matchLoc;
			minMaxLoc(result, &minVal, &maxVal, &minLoc, &maxLoc, Mat());
			matchLoc = maxLoc;
			matchLoc += searchWindow.tl();

			// Disable this tracker if its NCC is less than 0.95
			/*if (maxVal < 0.95) {
			t.setStateOutlier();
			continue;
			}*/

			// Updating references and displacement

			t.updateTL(matchLoc);

			// ==== End update

			//cout << t << endl;


			// If current tracker is an inlier (tracker with state "transition" should not count for global motion estimate)
			if (t.state == 1) {

				int voteWeight = 1;			// Default weight (for non context tracker)

				// Increase weights
				if (t.goodEstimate>9) {
					// Update the tracker weight
					voteWeight += t.goodEstimate / 10;			// Give more weight to current tracker if it had good performance in the past
					voteWeight = min(voteWeight, 5);			// Set an upper limit to prevent giving too much weight to a single tracker
				}

				
				// Add a weighted vote to the current bucket
				int currentVote = (hs[t.displacement.x / binSide][t.displacement.y / binSide] += voteWeight);

				// Update current best
				if (currentVote > maxVote) {
					maxVote = currentVote;
					bestDisplacement = t.displacement;
				}

				for (int i = 0; i < voteWeight; ++i){
					displacementX.push_back(t.displacement.x);
					displacementY.push_back(t.displacement.y);
				}
			
			}

			
			// Check for 100% overlapped trackers
			/*
				// Combine tl.x, tl.y and patch size to get a key
				string trackerKey = "X" + toString(t.tl.x) + "Y" + toString(t.tl.y) + "S" + toString(t.patchSize);
				if (trackerIndex.count(trackerKey)) {
					// Already exists a tracker in this exact position.. Disable!

					//Log::warning("Tracker Disabled!");

					// Keep the tracker with more weight and increase its weight
					auto& storedTracker = trackers[trackerIndex[trackerKey]];
					if (storedTracker.goodEstimate >= t.goodEstimate) {

						storedTracker.goodEstimate += t.goodEstimate;
						t.setStateOutlier();

					} else {

						t.goodEstimate += storedTracker.goodEstimate;
						storedTracker.setStateOutlier();
						trackerIndex[trackerKey] = i;
					}

				} else {
					// Add this tracker to the index
					trackerIndex[trackerKey] = i;
				}
			// === End overlap check
			*/
		}

		// If the clustering procedure did not have a good results, use the median
		if (maxVote < 40) {
			//cout << "Current votes: " << maxVote << endl;
			bestDisplacement.x = medianVector<int>(displacementX);
			bestDisplacement.y = medianVector<int>(displacementY);
		}

		int inlierTrackers = 0;
		int outlierTrackers = 0;
		int disabledTrackers = 0;
		// Update tracker state
		for (int i = 0; i<trackers.size(); i++) {

			auto& t = trackers[i];
			//cout << "["<<i<<"]"<< t << endl;

			// If current state is outlier no need to recheck it
			if (t.state != 2) {
				Point2i displacementError = t.displacement - bestDisplacement;

				// If current tracker displacement doesn't agree with the global mostion estimate, set the tracker as outlier
				if (abs(displacementError.x)>displacementErrorThreshold || abs(displacementError.y)>displacementErrorThreshold) {

					t.setStateOutlier(frameCount);
					++disabledTrackers;

					if (this->debug)
						drawRect(frame, Rect(t.tl.x,t.tl.y,t.patchSize,t.patchSize), Scalar(0,0,255), 1);

				}else{
					t.goodEstimate++;

					// If the current tracker is transition
					// and dispacement agrees with the global mostion estimate, set it as inlier after the validiation period
					if (t.state == 3) {

						// Each tracker can have a different transition threshold
						int currentTransitionThreshold = 
							t.transitionThreshold == -1 ? this->transitionThreshold : t.transitionThreshold;

						if (t.goodEstimate >= currentTransitionThreshold)
							t.setStateInlier();
					}
				}
			}

			if (t.state == 2)
				++outlierTrackers;
			else if (t.state == 1)
				++inlierTrackers;

		}


		// Update bounding box with the bestDisplacement and check its position
		rectUpdate(objectROI, bestDisplacement);
		rectInsideImage(currentFrame, objectROI);


		// Info about trackers
		Log(2) << "Trackers: " << trackers.size() << ", inlier: " << inlierTrackers << ", transition: " << (trackers.size() - inlierTrackers) - outlierTrackers << ", disabled now: " << disabledTrackers << endl;
		Log(2) << "BestDisplacement: " << bestDisplacement << ", with votes: " << maxVote << endl;

		
		float inlierPercentage = (float)inlierTrackers / trackers.size();

		// Critical situation, faster reinitialization
		if (inlierPercentage < 0.15f ) {

			//cout << "Too few inliers: "<<inlierTrackers << endl;
			disabledTrackersToTransitionIndipendently(1,1);

		}
		
		
		if (frameCount % 15 == 0 && inlierPercentage < 0.45f) {
			disabledTrackersToTransitionIndipendently();
		}
		
		
		/*
		// If % of all trackers are disabled, reinit them (or less than a threshold)
		if (inlierTrackers < 15) {				// If less than inlier trackers survived

		
			// Check if the previous transition was not too close
			if (frameCount < transitionDelay || counters.check("transition", frameCount, transitionDelay) //|| inlierTrackers < 10
			) {

				//cout << " Starting transition from outliers..."<<endl;

				counters.set("transition", frameCount);
				disabledTrackersToTransition(currentFrame, objectROI);

			}
		}
		*/

		if (this->debug)
			for (auto& t : trackers)	
				t.draw(frame);


		cv::swap(prevFrame, currentFrame);



		
		//printMultiMap<map<int, map<int, int>>>(hs);
		
		//showImage(currentFrame, "current frame");	
		
		//waitKey(0);
	}

	// Initialize multiple grids with different patch/cell sizes
	void initMultiSizeTrackers() {

	
		Log(1) << "Initialized with MultiSize Trackers" << endl;
		trackers.clear();
	
		

		// Automatic parameter setting
		if (trackerSizes.empty()) {

			int minRoiSide = min(objectROI.width, objectROI.height);
			int maxRoiSide = max(objectROI.width, objectROI.height);

			if (minRoiSide > 65) {			// Big Object

				cout << "Detected big object" << endl;

				//trackerSizes.push_back(18);	trackerSpaces.push_back(0);
				//trackerSizes.push_back(16);	trackerSpaces.push_back(-3);
				
				//trackerSizes.push_back(18);	trackerSpaces.push_back(-5);   // no failures for sphere and trellis, failures with motocross

				trackerSizes.push_back(18);	trackerSpaces.push_back(-5);
				trackerSizes.push_back(16);	trackerSpaces.push_back(1);	

			} else if (maxRoiSide > 80) {	// Elongated object
				
				cout << "Detected elongated object" << endl;

				//trackerSizes.push_back(10);	trackerSpaces.push_back(-5);
				//trackerSizes.push_back(18);	trackerSpaces.push_back(-9);

				trackerSizes.push_back(10);	trackerSpaces.push_back(-5);
				trackerSizes.push_back(16);	trackerSpaces.push_back(-3);
				trackerSizes.push_back(20);	trackerSpaces.push_back(-15);

		
			} else  {

				cout << "Detected small/standard object" << endl;

				trackerSizes.push_back(7);	trackerSpaces.push_back(2);
				trackerSizes.push_back(10);	trackerSpaces.push_back(-5);
				//trackerSizes.push_back(18);	trackerSpaces.push_back(-9);


				//trackerSizes.push_back(7);	trackerSpaces.push_back(0);
				//trackerSizes.push_back(10);	trackerSpaces.push_back(-2);
			}

		}

		int nTrackers = 0;
		for (auto i = 0; i < trackerSizes.size(); ++i) {


			auto currentPatchSize = trackerSizes[i];
			auto currentPatchSpace = trackerSpaces[i];

			if (
				currentPatchSize<3 ||						// Patch Size too small
				currentPatchSize + currentPatchSpace < 2	// Patches too overlapped
				) {
				continue;
			}

			int trackersRow = objectROI.height / (currentPatchSize + currentPatchSpace);
			int trackersCol = objectROI.width / (currentPatchSize + currentPatchSpace);


			Log(0) << " Size [" << currentPatchSize << "x" << currentPatchSize << "], Space ["<<currentPatchSpace<<"] TrackersRow: " << trackersRow << ", trackersCol: " << trackersCol << ", total: " << trackersCol*trackersRow << endl;

			for (int i = 0; i<trackersRow; ++i)
				for (int j = 0; j<trackersCol; ++j) {

					if (nTrackers >= 140)
						break;

					Point2i relativeTL((int)j*(currentPatchSize + currentPatchSpace), (int)i*(currentPatchSize + currentPatchSpace));

					Point2i absoluteP1(objectROI.x + relativeTL.x, objectROI.y + relativeTL.y);				// tl
					Point2i absoluteP2(objectROI.x + relativeTL.x + currentPatchSize, objectROI.y + relativeTL.y);	// sopra destra
					Point2i absoluteP3(objectROI.x + relativeTL.x + currentPatchSize, objectROI.y + relativeTL.y + currentPatchSize);	//br
					Point2i absoluteP4(objectROI.x + relativeTL.x, objectROI.y + relativeTL.y + currentPatchSize);

					// At least one corner of the patch must lie within the object rotated rect
					if (
						pointPolygonTest(this->corners, absoluteP1, false) >= 0 ||
						pointPolygonTest(this->corners, absoluteP2, false) >= 0 /*||
						pointPolygonTest(this->corners, absoluteP3, false) >= 0 ||
						pointPolygonTest(this->corners, absoluteP4, false) >= 0*/
					) {
						addTracker(currentFrame, objectROI, relativeTL, currentPatchSize);
						trackers.back().draw(frame);
						nTrackers++;
					}


			}
		}

	}


	// Add a new tracker
	void addTracker(const Mat& currentFrame, const Rect& objectROI, const Point2i& relativeTL, const int& patchSize, bool context = false){


		trackers.push_back(LocalTracker(
			relativeTL,
			patchSize,
			context
		));

		trackers.back().reinit(currentFrame, objectROI);

		//trackers.back().draw(frame);

	}


	// Reinit trackers with state outlier indipendetly ("outlier"->"transition")
	void disabledTrackersToTransitionIndipendently(int delay = -1, int threshold = -1) {

		if (delay < 0)
			delay = transitionDelay;

		if (threshold < 0)
			threshold = transitionThreshold;

		int reinitialized = 0;

		for (auto i = 0; i<trackers.size(); ++i) {
			// If the tracker has been outlier enough
			if (trackers[i].state == 2 && frameCount - trackers[i].outlierAt > delay) {

				int state = 3;
				// If we disabled transitionThreshold put it directly as inlier
				if (threshold == 0)
					state = 1;

				++reinitialized;
				trackers[i].reinit(currentFrame, objectROI, state, threshold);			// Set transition mode or directly inlier
			}
		}

		Log(2) << "Reinitialized: " << reinitialized << endl;
	}


	// Needs further testing
	void autoSetting(){

		vector<int> divisors = {3,4,7};

		int minRoiSide = min(objectROI.width, objectROI.height);
		int maxRoiSide = max(objectROI.width, objectROI.height);

		maxRoiSide = min(maxRoiSide, 60);

		for (auto d : divisors){
			trackerSizes.push_back( maxRoiSide / d );
			trackerSpaces.push_back( - ((maxRoiSide / d) / 2) );
		}

		trackerSpaces.pop_back();
		trackerSpaces.push_back(0);

	}


public:
	// ============
	//  Parameters
	// ============

	int patchEnlargeWindow = 20;		// Search window [default: 20], should be setted in according to the object/camera speed

	int displacementErrorThreshold = 7;	// Max error between the tracker displacement and the best displacement to become outlier
	int binSide = 1;					// Accumulator space granularity

	int transitionThreshold = 3;		// Validation period to be promoted to inlier state [default: 3]
	int transitionDelay = 20;			// Delay before reinitializing an outlier tracker

	// If empty, they will be automatically setted
	vector<int> trackerSizes;			// Defines the sizes of the patches used for block matching
	vector<int> trackerSpaces;			// Defines the space if positive (the overlap if negative) between patches

private:
	Mat prevFrame;

	
	vector<LocalTracker> trackers;

};
