

// Logger Class
/*
	- Example
		Log(level) << "Info: "<<endl;

	- Levels
	    0 - one-time information (i.e. initialization)
		1 - important (default)
		2 - semi important
		3 - other

	- Change minimum level with:
		Log::setLevel(2);



	//std::ofstream out("out.txt");
	//std::cout.rdbuf(out.rdbuf()); //redirect std::cout to out.txt!
*/

class Log{
public:
	Log(const int& msgLevel = 1){
		this->msgLevel = msgLevel;
	}

	template <class T>
	Log& operator<<(const T& v){
		if (msgLevel<=level) {
			std::cout << v;
		}
		return *this;
	}

	// For manipulators like endl
	Log& operator<< (ostream& (*pf) (ostream&)) {
		if (msgLevel<=level) {
			cout << pf;
		}
			
		return *this;
	}

	static void setLevel(const int& newLevel) {
		level = newLevel;
	}



	// Fatal
	static void fatal(const string& err) {
		string msg = "[FATAL] "+err;
		cout << endl <<"======="<<endl<<msg<<endl<<"======="<<endl;
		sp();
		throw runtime_error(msg);
	}

	// Non-fatal warning
	static void warning(const string& msg) {
		cout << endl <<"========="<<endl<<"[WARNING] "<<msg<<endl<<"========="<<endl;
	}
			
		


	~Log(){}

	static int level;
	int msgLevel;
};

#ifdef _DEBUG
	int Log::level = 0;
#else
	int Log::level = 0;
#endif





class Timer{
	public:

	Timer() {
		totalSec = 0;
		start = clock();
	}

	void timerStart(){
		start = clock();
	}

	double timerEnd(string msg="default", bool append = false) {	

		clock_t current = clock();
		double sec = ((double)current - start) / CLOCKS_PER_SEC;

		if (append)
			totalSec += sec;
		else
			cout << "== TIMER ["<<msg<<"] took: "<<sec<<" sec"<<endl;

		return sec;
	}

	void print(string msg="default") {
		cout << "== TIMER ["<<msg<<"] took total: "<<totalSec<<" sec"<<endl;
	}

	private:
		clock_t start;
		double totalSec;

};



// Manage video output frame by frame
/* Example:
		VideoMaker myVideoMaker("./path/");
		myVideoMaker.writeFrame(frameImage);

*/
class VideoMaker {

	public:
	VideoMaker(string writePath = "./") {
		setWritePath(writePath);
	}
	~VideoMaker(){}

	// Add a frame to the current video
	void writeFrame(const Mat& frame, bool writeSingleImage = false) {
		frameCount++;
		if (frameCount==1) {
			capOutput = VideoWriter(
				writePath + "trackingOutput.avi",
				CV_FOURCC('D','I','V','X'),
				20,
				frame.size()
			);

			if (!capOutput.isOpened()) {
				Log::fatal("Output video could not be opened, check path: " + writePath);
			}
		}


		capOutput.write(frame);

		if (writeSingleImage) {
			imwrite(writePath + "frame/" + toString(frameCount) + ".png",frame);
		}

	}

	void setWritePath(string writePath) {
		this->writePath = writePath;
	}

	private:
		VideoWriter capOutput;
		int frameCount = 0;
		string writePath;
};


// Manage ground truth
// To output the tracker bounding boxes just call .setOutputPath("trackerOutput.txt") 
class GTManager {

	public:
			
		float overlapThreshold = 0.5;
		
		Rect boundingBoxGT;

		static const int FORMAT_WH = 0;			// x,y,width,height
		static const int FORMAT_TLBR = 1;		// x,y,x,y
		static const int FORMAT_VOTPOLY = 2;	// x1,y2,...

		int currentFormat = 0;		

		GTManager(){}

		void setFormat(int f) {
			this->currentFormat = f;
		}

		Rect getNextRect() {
			
			char ch;

			if ( currentFormat == FORMAT_WH) {
				float x, y, width, height;
				groundTruth >> x >> ch >> y >> ch >> width >> ch >> height;

				boundingBoxGT = Rect((int)x, (int)y, (int)width, (int)height);
			} else if (currentFormat == FORMAT_TLBR){

				float x1, y1, x2, y2;
				groundTruth >> x1 >> ch >> y1 >> ch >> x2 >> ch >> y2;

				boundingBoxGT = Rect(Point2f(x1, y1), Point2f(x2, x2));
			}
			else if (currentFormat == FORMAT_VOTPOLY) {
				
				float x1, y1, x2, y2, x3, y3, x4, y4;
				groundTruth
					>> x1 >> ch >> y1 >> ch
					>> x2 >> ch >> y2 >> ch
					>> x3 >> ch >> y3 >> ch
					>> x4 >> ch >> y4;

				corners.clear();
				corners.push_back(Point2f(x1, y1));
				corners.push_back(Point2f(x2, y2));
				corners.push_back(Point2f(x3, y3));
				corners.push_back(Point2f(x4, y4));
				boundingBoxGT = boundingRect(corners);
			}
	
			if (boundingBoxGT.area()) {
				objectFrames++;				// Frames in which the object is visible
			} 

			this->totalFrames++;

			return boundingBoxGT;
		}

		float calcOverlap(const Rect& boundingBox) {

			float overlap = 0.0f;

			// If the object is visible in the ground truth
			if (boundingBoxGT.area()>5) {

				Rect rIntersect(boundingBoxGT&boundingBox);				
				int intersect = rIntersect.area();
				int unionArea = boundingBoxGT.area()+boundingBox.area();

				overlap = (float) intersect / ( unionArea - intersect );
				overlaps.push_back(overlap);

				if (overlap>=overlapThreshold)
					correctFrames++;

			}

			

			writeBoundingBox(boundingBox);

			return overlap;

		}

		float calcMeanOverlap() {
				
			return meanVector<float>(this->overlaps);
		}

			
		void writeBoundingBox(const Rect& bbox) {
			
			if (!outputFile.empty()) {

				int x1,x2,x3,x4;

				if ( this->currentFormat == FORMAT_TLBR ) {

					x1 = bbox.x;
					x2 = bbox.y;
					x3 = bbox.br().x;
					x4 = bbox.br().y;

				} else {

					x1 = bbox.x;
					x2 = bbox.y;
					x3 = bbox.width;
					x4 = bbox.height;

				}

				if (bbox.area()<5)
					writeNaN();
				else
					outputStream << x1 << "," << x2 << "," << x3 << "," << x4 << endl;
			}
		}

		void writeNaN() {
			outputStream << "NaN,NaN,NaN,NaN" << endl;
		}

			
		void open(const string& file) {
			groundTruthFile = file;
			if (!file.empty()) {
				groundTruth.open(this->groundTruthFile);
				if (!groundTruth.is_open())
					Log::fatal("Groundtruth could not be opened ["+(this->groundTruthFile)+"].");
			}

		}

		void setOutputFile(const string& file) {
			
			outputFile = file;
				
			outputStream.open(file);
			if (!outputStream.is_open())
				Log::fatal("Output file could not be opened ["+(this->outputFile)+"].");
			
		}

		float printStats() {
			float meanAccuracy=0.0f;
			meanAccuracy = calcMeanOverlap();
			cout << endl<<endl<<"====="<<endl<<"STATS"<<endl<<"====="<<endl<<"Mean accuracy: "<<meanAccuracy<<endl;
			cout << "Recall: "<<(float)this->correctFrames/objectFrames<<endl
				 <<"Correct Frame: "<<this->correctFrames
				 <<", out of: "<<objectFrames<<" with object, total frames: "<<totalFrames<<endl;


			return meanAccuracy;
		}

		vector<Point2f> getCorners(){

			if (corners.empty()) {
				corners = rectToCorners(boundingBoxGT);
				/*
				corners.push_back(Point2f(boundingBoxGT.x, boundingBoxGT.y));
				corners.push_back(Point2f(boundingBoxGT.br().x, boundingBoxGT.br().y));
				corners.push_back(Point2f(boundingBoxGT.x + boundingBoxGT.width, boundingBoxGT.y));
				corners.push_back(Point2f(boundingBoxGT.x, boundingBoxGT.y + boundingBoxGT.height));
				*/
			} 


			return corners;
		}
		
	private:
		string groundTruthFile;
		ifstream groundTruth;

		vector<Point2f> corners;	// In case of a rotated rect
		vector<float> overlaps;
		ofstream outputStream;
		string outputFile;
		int correctFrames = 0;		// Frames where the tracker bounding box overlap with gt is > threshold
		int totalFrames = 0;
		int objectFrames = 0;		// Frames where the object is visible based on the ground truth

		

};

// Example:
/*
	if ( frameCount - counters.get("transition") > 20 ) {
		// do 
		counters.set("transition",frameCount);
	}

	// OR equivalent:

	if ( counters.check("transition", frameCount, 20) ) {
	    // do 

		counters.set("transition",frameCount);
	}
*/
class Counter {

public:
	Counter(){}
	void set(string name, int value){
		
		counters[name] = value;
	}

	int get(string name){

		return counters[name];
	}

	bool check(string name, int compareAgainst, int threshold) {

		return (compareAgainst - counters[name] > threshold);
	}

private:
	map<string, int> counters;

};

// To emulate VOT class way of reading with imread
class VideoCapture2{

public:
	void open(string openPath) {
		this->path = openPath;
	}

	bool read(Mat& im){

		char buff[1000];
		snprintf(buff, i, path.c_str());
		string currentFrame = buff;

		//cout << "Reading: " << currentFrame << endl;

		//im = cv::imread(currentFrame, CV_LOAD_IMAGE_COLOR);
		im = cv::imread(currentFrame, CV_LOAD_IMAGE_GRAYSCALE);

		i++;

		// If the sequence starts from 0 or 1
		if (i == 1 && im.empty())
			read(im);

		return true;
	}


private:
	int i = 0;
	string path;

};

// Abstract class for a tracker 
class Tracker {


public:
	Tracker(){}
	~Tracker(){ 
		destroyAllWindows(); 
	}

	void openDataset(string path, string images = "%08d.jpg", string gtName = "groundtruth.txt", int format = GTManager::FORMAT_VOTPOLY) {
		cap.open(path + images);
		gt.open(path + gtName);
		gt.setFormat(format);

		//capOutput.setWritePath("./");
		gt.setOutputFile("./output.txt");

		Log(0) << "INITIALIZATION on dataset: " << path << endl;
	}

	// Simple implementation of run 
	float run(int startFrom = 0) {
		t.timerStart();
		int current = 0;
		while (1) {
			this->cap.read(frame);
			
			// End of sequence?
			if (frame.empty())
				break;


			Rect gtBB = gt.getNextRect();

			++current;
			if (startFrom > current)
				continue;

			if (frameCount == 0) {
				firstBB = gtBB;
				Log(0) << "FirstBB: " << firstBB << ", frame size: " << frame.size() << endl<<endl;
				corners = gt.getCorners();
				
				//drawRect(frame, corners);
				//showImage(frame, "Corners");
				//waitKey(0);
			}

			Log(1) << endl << endl << "Frame #" << frameCount << endl;

			// Track & show frame
			this->track();
			this->showFrame();

			
			gt.calcOverlap(objectROI);

			//capOutput.writeFrame(frame,true);
			//gt.writeBoundingBox(objectROI);

			// Keyboard shortcut
			if (!keyboardManagerStandard(1))
				break;
		}
		//return 0.0f;
		return this->printStats();
	}

	// Prints: FPS, recall and accuracy
	float printStats(){
		double sec = t.timerEnd();
		cout << endl <<endl << "====STATS" << endl << "Frames per second: " << frameCount / sec << ", frames: " << frameCount << endl;
		float meanAccuracy = 0.0f;
		meanAccuracy = gt.printStats();
		cout << "=========" << endl << endl << endl;

		return meanAccuracy;
	}

	// Shows the current frame
	void showFrame() {

		if (isObjectFound) {
			drawRect(frame, objectROI,Scalar(0,255,0),4);
			//drawCross(frame, objectROI);

		} else {
			// White rect
			drawRect(frame, objectROI, Scalar(255,255,255));
		}

		showImage(frame, "Tracker");
	}

	// Prepare for tracking (should be called in the .track() method)
	void prepare(){
		++frameCount;

		if (frame.channels() > 1) {
			cvtColor(frame, currentFrame, COLOR_BGR2GRAY);
		} 
		else {
			frame.copyTo(currentFrame);
			cvtColor(frame, frame, CV_GRAY2RGB);						// Frame should be color for drawing functions
		}
	}

	// Each specific tracker should implement its track method
	virtual void track() = 0;


public:
	// Processing
	Mat frame, currentFrame;
	Counter counters;
	int frameCount = 0;

	// Other
	Timer t;
	bool debug = true;
	VideoMaker capOutput;

	// Initialization/input-related
	GTManager gt;
	VideoCapture2 cap;
	Rect firstBB;
	vector<Point2f> corners;		// In case of a rotated rect

	// Object-related
	bool isObjectFound = true;
	Rect objectROI;

};
