/* 
 *  Author : Tomas Vojir
 *  Date   : 2013-06-05
 *  Desc   : Simple class for parsing VOT inputs and providing 
 *           interface for image loading and storing output.
 */ 

#ifndef CPP_VOT_H
#define CPP_VOT_H

#include <string>
#include <fstream>
#include <iostream>
#include "opencv2/opencv.hpp"

#include "trax.h"

// Bounding box type
typedef struct {
    float x1;
    float y1;
    float x2;
    float y2;
    float x3;
    float y3;
    float x4;
    float y4;
} VOTPolygon;

class VOT
{
public:
    VOT(const std::string & region_file, const std::string & images, const std::string & ouput)
    {
        p_region_stream.open(region_file.c_str());
        VOTPolygon p;
        if (p_region_stream.is_open()){
            char ch;
            p_region_stream >> p.x1 >> ch >> p.y1 >> ch;
            p_region_stream >> p.x2 >> ch >> p.y2 >> ch;
            p_region_stream >> p.x3 >> ch >> p.y3 >> ch;
            p_region_stream >> p.x4 >> ch >> p.y4;

        }else{
            std::cerr << "Error loading initial region in file " << region_file << "!" << std::endl;
            p.x1=0;
            p.y1=0;
            p.x2=0;
            p.y2=0;
            p.x3=0;
            p.y3=0;
            p.x4=0;
            p.y4=0;
        }

        p_init_polygon = p;

        p_images_stream.open(images.c_str());
        if (!p_images_stream.is_open())
            std::cerr << "Error loading image file " << images << "!" << std::endl;

       
        p_output_stream.open(ouput.c_str());
        if (!p_output_stream.is_open())
            std::cerr << "Error opening output file " << ouput << "!" << std::endl;
    }

    ~VOT()
    {
        p_region_stream.close();
        p_images_stream.close();
        p_output_stream.close();
    }

    inline VOTPolygon getInitPolygon() const 
    {   return p_init_polygon;    }

	// Converts VOTPolygon to the standard cv::Rect
	inline cv::Rect getInitRectangle() const {
		vector<Point2f> points;
		points.push_back(Point2f(p_init_polygon.x1, p_init_polygon.y1));
		points.push_back(Point2f(p_init_polygon.x2, p_init_polygon.y2));
		points.push_back(Point2f(p_init_polygon.x3, p_init_polygon.y3));
		points.push_back(Point2f(p_init_polygon.x4, p_init_polygon.y4));
		return boundingRect(points);
	}

	// Returns vector of points of the corners of the init rotated rect
	inline std::vector<cv::Point2f> getInitCorners() const {
		vector<Point2f> points;
		points.push_back(Point2f(p_init_polygon.x1, p_init_polygon.y1));
		points.push_back(Point2f(p_init_polygon.x2, p_init_polygon.y2));
		points.push_back(Point2f(p_init_polygon.x3, p_init_polygon.y3));
		points.push_back(Point2f(p_init_polygon.x4, p_init_polygon.y4));
		return points;
	}

    inline void outputPolygon(const VOTPolygon & poly)
    {
      p_output_stream << poly.x1 << ", " << poly.y1 << ", ";
      p_output_stream << poly.x2 << ", " << poly.y2 << ", ";
      p_output_stream << poly.x3 << ", " << poly.y3 << ", ";
      p_output_stream << poly.x4 << ", " << poly.y4 << std::endl;
    }

	// Converts cv::Rect to the VOTPolygon and calls outputPolygon
	inline void outputBoundingBox(const cv::Rect& bbox) {

		VOTPolygon poly;

		/*
		// P1 = bottom left corner
		poly.x1 = bbox.x;
		poly.y1 = bbox.y + bbox.height;

		// P2 = top left corner
		poly.x2 = bbox.x;
		poly.y2 = bbox.y;

		// P3 = top right corner
		poly.x3 = bbox.x + bbox.width;
		poly.y3 = bbox.y;

		// P4 = bottom right corner
		poly.x4 = bbox.x + bbox.width;   // or simply bbox.br();
		poly.y4 = bbox.y + bbox.height;
		*/


		// ncc.cpp style

		poly.x1 = bbox.x;
		poly.y1 = bbox.y;
		poly.x2 = bbox.x + bbox.width;
		poly.y2 = bbox.y;
		poly.x3 = bbox.x + bbox.width;
		poly.y3 = bbox.y + bbox.height;
		poly.x4 = bbox.x;
		poly.y4 = bbox.y + bbox.height;

		this->outputPolygon(poly);

	}

    inline int getNextImage(cv::Mat & img)
    {
    if (p_images_stream.eof() || !p_images_stream.is_open())
            return -1;

    std::string line;
    std::getline (p_images_stream, line);
    //std::cout << line << std::endl;
	img = cv::imread(line, CV_LOAD_IMAGE_GRAYSCALE);
	//img = cv::imread(line, CV_LOAD_IMAGE_COLOR);

	//VideoCapture v;		// VideoCapture seems better for reading image than imread
	//v.open(line);
	//v.read(img);

    return 1;
  }

private:
    VOTPolygon p_init_polygon;
    std::ifstream p_region_stream;
    std::ifstream p_images_stream;
    std::ofstream p_output_stream;

};

#endif //CPP_VOT_H
